var NewSpotlightSearchHandler = (function(options) {
	var _inputElement  = _getElement(_getOption("inputSelector", true));
	var _pickListElement = _getElement(_getOption("pickListSelector", true));
	var _pickListElementIsSelect = _pickListElement.tagName === "SELECT";
	var _rowSourceAttrName = _getOption("rowSourceAttrName", true);
	var _searchFields = _getOption("searchFields");
	var _onSelectionMade = _getOption("onSelectionMade");
	var _onReSortAfterFilter_compareItems = _getOption("onReSortAfterFilter_compareItems");
	var _onAfterFilter = _getOption("onAfterFilter");
	var _onSelectedIndexChanged;
	var _oldDocumentKeyDown;
	var _items;

	_refresh();
	_filter();

	_inputElement.addEventListener("input", function(event) {
		_filter();
	});

	_oldDocumentKeyDown = document.onkeydown;
	document.onkeydown = _documentKeyDown;

	if (_pickListElementIsSelect) {
		_inputElement.addEventListener("keyup", _inputElementKeyUp);
		_pickListElement.addEventListener("change", _selectedIndexChanged);
		_pickListElement.addEventListener("keydown", _pickListElementKeyDown);
		_pickListElement.addEventListener("keyup", _pickListElementKeyUp);
		document.addEventListener("dblclick", _documentDblClick);
	}

	_onSelectedIndexChanged = _getOption("onSelectedIndexChanged");
	if (_onSelectedIndexChanged)
		setTimeout(_onSelectedIndexChanged, 1);

	// ========================================================================
	function _contentUpdated() {
		if (!_pickListElementIsSelect)
			return;

		if (_pickListElement.selectedIndex < 0 && _pickListElement.children.length > 0 && $(_inputElement).val()) {
			_pickListElement.selectedIndex = 0;
		}
		_selectedIndexChanged();
	}

	function _documentDblClick(event) {
		if (event.target.tagName === "OPTION" && event.target.parentNode === _pickListElement)
			_selectionMade();
	}

	function _documentKeyDown(event) {
		try {
			switch (event.which) {
				case 27: // escape
					var activeElement = document.activeElement;
					if (activeElement === _inputElement) {
						var input = $(_inputElement);
						if (input.val()) {
							input.val("");
							_filter();
							return false;
						}
					} else if (activeElement === _pickListElement) {
						_inputElement.focus();
						return false;
					}
					break;
				default:
					break;
			} // switch
			if (_oldDocumentKeyDown)
				return _oldDocumentKeyDown(event);
			return true;
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.documentKeyDown");
			return false;
		}
	}

	function _filter() {

		function SearchTermMatches(searchText) {
			this.terms =
				searchText
					.split(/[" "]+/)
					.map(function(term) {
						return {
							matched: false,
							term: term
						};
					});

			this.resetMatches = function() {
				for (var i = 0; i < this.terms.length; i++) {
					this.terms[i].matched = false;
				}
			};

			this.allTermsMatched = function() {
				for (var i = 0; i < this.terms.length; i++) {
					if (!this.terms[i].matched)
						return false;
				}
				return true;
			};

			this.getUnmatchedTerms = function() {
				return this.terms.filter(function(item) {
					return !item.matched;
				});
			};
		}

		function checkStringForMatches(str, searchTerms) {
			if (!str)
				return;

			str = str.toLowerCase();
			var unmatchedTerms = searchTerms.getUnmatchedTerms();
			unmatchedTerms.forEach(function(unmatchedTerm) {
				if (str.indexOf(unmatchedTerm.term) >= 0) {
					unmatchedTerm.matched = true;
				}
			});
		}

		function itemMatches(item, searchTerms) {
			searchTerms.resetMatches();

			if (!_searchFields || _searchFields.length === 0) {
				checkStringForMatches(item, searchTerms);
				return searchTerms.allTermsMatched();
			}

			for (var i = 0; i < _searchFields.length && !searchTerms.allTermsMatched(); i++) {
				var searchField = _searchFields[i];
				var field = item[searchField];
				if (Array.isArray(field)) {
					for (var j = 0; j < field.length && !searchTerms.allTermsMatched(); j++) {
						checkStringForMatches(field[j], searchTerms);
					}
				} else {
					checkStringForMatches(item[searchField], searchTerms);
				 }
			}
			return searchTerms.allTermsMatched();
		}

		function getMatches(searchText) {
			if (!searchText)
				return _items;
			searchText = searchText.toLowerCase();
			var searchTerms = new SearchTermMatches(searchText);
			return _items.filter(function(item) {
				return itemMatches(item.rowSource, searchTerms);
			});
		}

		// ===== _filter =======================================
		try {
			var searchText = _inputElement.value;
			var matches = getMatches(searchText);
			if (matches.length === _items.length && matches.length === _pickListElement.children.count)
				return;

			$(_pickListElement).empty();

			if (_onReSortAfterFilter_compareItems) {
				var firstCompare = true;
				matches.sort(function(a, b) {
					var result = _onReSortAfterFilter_compareItems(a.rowSource, b.rowSource, searchText, firstCompare);
					if (firstCompare)
						firstCompare = false;
					return result;
				});
			}

			matches.forEach(function(item) {
				$(_pickListElement.appendChild(item.row)).data(_rowSourceAttrName, item.rowSource);
			});

			_contentUpdated();

			if (_onAfterFilter)
				_onAfterFilter(searchText);
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.filter");
		}
	} // _filter

	function _getElement(selector) {
		var result = document.querySelector(selector);
		if (!result)
			throw Error("Element for selector '" + select + "' not found");
		return result;
	}

	function _getItemAtIndex(index) {
		if (!_pickListElementIsSelect)
			throw Error("getItemAtIndex only valid on SELECT elements");
		try {
			if (index < 0 || index >= _pickListElement.children.length)
				return null;
			return _getRowSource(_pickListElement.children[index]);
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.getItemAtIndex");
			return null;
		}
	}

	function _getOption(propertyName, required) {
		var result = options[propertyName];
		if (result || !required)
			return result;
		throw Error("Required option '" + propertyName + "' was not specified");
	}

	function _getRowSource(row) {
		var rowSource = $(row).data(_rowSourceAttrName);
		/* jshint -W041 */
		if (rowSource == null)
			throw Error("Could not find row source attribute '" + _rowSourceAttrName + "'");
		/* jshint +W041 */
		return rowSource;
	}

	function _getSelectedItem() {
		if (!_pickListElementIsSelect)
			throw Error("getSelectedItem only valid on SELECT elements");
		try {
			var i = _pickListElement.selectedIndex;
			if (i < 0)
				return null;
			return _getRowSource(_pickListElement.children[i]);
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.getSelectedItem");
			return null;
		}
	}

	function _inputElementKeyUp(event) {
		try {
			var selectedIndex;
			var children;

			switch (event.which) {
				case 13: // enter
					_selectionMade();
					break;
				case 38: // up arrow
					selectedIndex = _pickListElement.selectedIndex;
					children = _pickListElement.children;
					if (children.length > 0) {
						if (selectedIndex <= 0) {
							_pickListElement.selectedIndex = children.length - 1;
						} else {
							_pickListElement.selectedIndex = selectedIndex - 1;
						}
						_selectedIndexChanged();
						_pickListElement.focus();
					}
					break;
				case 40: // down arrow
					selectedIndex = _pickListElement.selectedIndex;
					children = _pickListElement.children;
					if (children.length > 0) {
						if (selectedIndex < 0 || selectedIndex === children.length-1) {
							_pickListElement.selectedIndex = 0;
						} else {
							_pickListElement.selectedIndex = selectedIndex + 1;
						}
						_selectedIndexChanged();
						_pickListElement.focus();
					}
					break;
				default:
					break;
			}
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.inputElementKeyUp");
		}
	}

	_lastTimeWhenPickListUpArrowKeyDown = 0;
	_lastSelectedIndexWhenPickListUpArrowKeyDown = -1;
	_repeatingPickListUpArrow = false;
	function _pickListElementKeyDown(event) {
		try {
			switch (event.which) {
				case 38: // up arrow
					_repeatingPickListUpArrow = Date.now() - _lastTimeWhenPickListUpArrowKeyDown < 100;
					_lastTimeWhenPickListUpArrowKeyDown = Date.now();
					_lastSelectedIndexWhenPickListUpArrowKeyDown = _pickListElement.selectedIndex;
					break;
				default:
					_repeatingPickListUpArrow = false;
					break;
			}
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.pickListElementKeyDown");
		}
	}

	function _pickListElementKeyUp(event) {
		try {
			var selectedIndex;
			var children;

			switch (event.which) {
				case 13: // enter
					_selectionMade();
					break;
				case 38: // up arrow
					selectedIndex = _pickListElement.selectedIndex;
					if (selectedIndex === 0 &&
						_lastSelectedIndexWhenPickListUpArrowKeyDown === 0 &&
						!_repeatingPickListUpArrow)
						_inputElement.focus();
					break;
				default:
					break;
			}
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.pickListElementKeyUp");
		}
	}

	function _refresh() {
		try {
			_items = [];
			var children = _pickListElement.children;
			for (var i = 0; i < children.length; i++) {
				var child = children[i];
				var rowSource = _getRowSource(child);
				var item = { row: child, rowSource: rowSource };
				_items.push(item);
			}
		} catch (e) {
			ErrorHandler.handle(e, "SearchHandler.refresh");
		}
	}

	function _selectedIndexChanged() {
		if (_onSelectedIndexChanged)
			_onSelectedIndexChanged();
	}

	function _selectionMade() {
		if (_onSelectionMade && _getSelectedItem())
			_onSelectionMade();
	}

	return {
		filter: function() {
			_filter();
		},

		getItemAtIndex: function(index) {
			return _getItemAtIndex(index);
		},

		getSelectedItem: function() {
			return _getSelectedItem();
		}
	};
});
